/**
 * Stinger is an HTTP Request Validation Engine
 * Copyright (C) 2006  Aspect Security, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact us at info@aspectsecurity.com or www.aspectsecurity.com
 *
 */

package org.owasp.stinger;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.dom.DOMSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.owasp.stinger.Category;
import org.owasp.stinger.Severity;
import org.owasp.stinger.rules.Rule;
import org.owasp.stinger.rules.RuleSet;

public class LearnerFilter implements Filter {
	
	public final static String STINGER_DEFAULT = "STINGER_DEFAULT";
	
	private static Category missing = null;
	
	private static Category malformed = null;
	
	private static String config = null;
	
	private static RuleSet ruleSet = null;
	
	public void init(FilterConfig filterConfig) {
		/** Pull template config location from Filter init parameter **/
		config = filterConfig.getInitParameter("config");
		ruleSet = new RuleSet(config);
		
		/** Build Default Categories **/
		missing = new Category(Category.MISSING, Severity.IGNORE);
		malformed = new Category(Category.MALFORMED, Severity.CONTINUE);
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
		HttpServletRequest hRequest = null;
		HttpServletResponse hResponse = null;
		
		if(request instanceof HttpServletRequest && response instanceof HttpServletResponse) {
			hRequest = (HttpServletRequest)request;
			hResponse = (HttpServletResponse)response;
			
			System.out.println("LearnerFilter: requested uri " + hRequest.getRequestURI());
			
			buildRulesFromRequest(hRequest);
			
			chain.doFilter(hRequest, hResponse);
		}
	}
	
	public void destroy() {
		writeRulesToConfig(config);
	}
	
	private synchronized void buildRulesFromRequest(HttpServletRequest request) {
		String uri = null;
		String name = null;
		String value = null;
		Enumeration e = null;
		Rule rule = null;
		
		uri = request.getRequestURI();
		e = request.getParameterNames();
		
		while(e.hasMoreElements()) {
			name = (String)e.nextElement();
			value = (String)request.getParameter(name);
			
			if(!hasRule(uri, name)) {
				/** No rule exists, create... **/
				rule = new Rule();
				rule.setName(name);
				rule.setPattern(getPattern(value));
				rule.setMissing(missing);
				rule.setMalformed(malformed);
				
				ruleSet.addParameterRule(uri, rule);
			} else {
				/** Rule exists, adjust pattern accordingly **/
				rule = ruleSet.getParameterRule(uri, name);
				adjustPattern(value, rule);
			}
		}
	}
	
	private boolean hasRule(String uri, String name) {
		boolean result = true;
		Rule rule = null;
		
		rule = ruleSet.getParameterRule(uri, name);
		
		if(rule.getName().equals(STINGER_DEFAULT)) {
			result = false;
		}
		
		return result;
	}
	
	private String getPattern(String value) {
		StringBuffer buffer = new StringBuffer();
		char[] array = value.toCharArray();
		HashSet<Character> set = new HashSet<Character>();
		
		/** Start with the safe default **/
		buffer.append("^[a-zA-Z0-9");
		
		/** Iterator over char's and append appropriately **/
		for(int i=0; i<array.length; i++) {
			if(!Character.isLetter(array[i]) && !Character.isDigit(array[i])) {
				if(!set.contains(array[i])) {
					buffer.append(array[i]);
					set.add(array[i]);
				}
			}
		}
		
		/** Close off regex **/
		buffer.append("]+$");
		
		return buffer.toString();
	}
	
	private void adjustPattern(String value, Rule rule) {
		String pattern = rule.getPattern();
		char[] array = value.toCharArray();
		HashSet<Character> set = new HashSet<Character>();
		
		/** Iterator over char's and replace appropriately **/
		for(int i=0; i<array.length; i++) {
			if(!Character.isLetter(array[i]) && !Character.isDigit(array[i])) {
				if(!set.contains(array[i])) {
					pattern.replace(array[i] + "]+$", "]+$");
					set.add(array[i]);
				}
			}
		}
		
		rule.setPattern(pattern);
	}
	
	private synchronized void writeRulesToConfig(String config) {
		String uri = null;
		String name = null;
		Rule rule = null;
		Document d = parseXmlFile(config);
		Element root = d.getDocumentElement();
		Element setElement = null;
		Element ruleElement = null;
		Element catElement = null;
		Element actsElement = null;
		Element actElement = null;
		Element pElement = null;
		Element e = null;
		HashMap<String, HashMap<String,Rule>> pRules = ruleSet.getParameterRules();
		HashMap<String, Rule> rules = null;
		Iterator uriItr = pRules.keySet().iterator();
		Iterator pItr = null;
		
		while(uriItr.hasNext()) {
			uri = (String)uriItr.next();
			setElement = d.createElement("ruleset");
			
			e = d.createElement("name");
			e.appendChild(d.createTextNode("Generated By Learner"));
			setElement.appendChild(e);
			
			e = d.createElement("path");
			e.appendChild(d.createTextNode(uri));
			setElement.appendChild(e);
			
			rules = pRules.get(uri);
			
			pItr = rules.keySet().iterator();
			
			while(pItr.hasNext()) {
				name = (String)pItr.next();
				rule = rules.get(name);
				
				ruleElement = d.createElement("rule");
				
				e = d.createElement("name");
				e.appendChild(d.createTextNode(rule.getName()));
				ruleElement.appendChild(e);
				
				e = d.createElement("regex");
				e.appendChild(d.createTextNode(rule.getPattern()));
				ruleElement.appendChild(e);
				
				catElement = d.createElement("missing");
				
				e = d.createElement("severity");
				e.appendChild(d.createTextNode(rule.getMissing().getSeverity().toString()));
				catElement.appendChild(e);
				
				ruleElement.appendChild(catElement);
				
				catElement = d.createElement("malformed");
				
				e = d.createElement("severity");
				e.appendChild(d.createTextNode(rule.getMalformed().getSeverity().toString()));
				catElement.appendChild(e);
				
				actsElement = d.createElement("actions");
				
				actElement = d.createElement("action");
				actElement.setAttribute("class", "org.owasp.stinger.actions.Log");
				
				pElement = d.createElement("parameter");
				
				e = d.createElement("name");
				e.appendChild(d.createTextNode("log"));
				pElement.appendChild(e);
				
				e = d.createElement("value");
				e.appendChild(d.createTextNode("default.log"));
				pElement.appendChild(e);
				
				
				actElement.appendChild(pElement);
				
				pElement = d.createElement("parameter");
				
				e = d.createElement("name");
				e.appendChild(d.createTextNode("level"));
				pElement.appendChild(e);
				
				e = d.createElement("value");
				e.appendChild(d.createTextNode("INFO"));
				pElement.appendChild(e);
				
				
				actElement.appendChild(pElement);
				
				pElement = d.createElement("parameter");
				
				e = d.createElement("name");
				e.appendChild(d.createTextNode("message"));
				pElement.appendChild(e);
				
				e = d.createElement("value");
				e.appendChild(d.createTextNode("parameter %name with value %value from %ip has been sanitized"));
				pElement.appendChild(e);
				
				
				actElement.appendChild(pElement);
				
				actsElement.appendChild(actElement);
				
				actElement = d.createElement("action");
				actElement.setAttribute("class", "org.owasp.stinger.actions.Sanitize");
				
				actsElement.appendChild(actElement);
				
				catElement.appendChild(actsElement);
				
				ruleElement.appendChild(catElement);
				
				setElement.appendChild(ruleElement);
			}
			
			/** We do not wish to write out another copy of the Stinger Default Rule **/
			if(!getField(setElement, "path").equals(STINGER_DEFAULT)) {
				root.appendChild(setElement);	
			}
		}
		
		/** Document Fully Built, Write to Disk **/
		writeXmlFile(config, d);
	}
	
	private Document parseXmlFile(String fileName) {
		Document doc = null;
		DocumentBuilderFactory bf = null;
		
		try {
			bf = DocumentBuilderFactory.newInstance();
			doc = bf.newDocumentBuilder().parse(new File(fileName));
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return doc;
	}
	
	private void writeXmlFile(String fileName, Document d) {
		File xmlOutputFile = new File(fileName);
		FileOutputStream fos = null;
		Transformer transformer = null;
		TransformerFactory factory = null;
		DOMSource source = new DOMSource(d);
		StreamResult result = null;
		
		try {
			fos = new FileOutputStream(xmlOutputFile);
			result = new StreamResult(new OutputStreamWriter(fos, "utf-8"));
			
			factory = TransformerFactory.newInstance();
			factory.setAttribute("indent-number", new Integer(4));
			transformer = factory.newTransformer();
			transformer.setOutputProperty("indent", "yes");
			transformer.transform(source, result);
		} catch (Exception e) {
			e.printStackTrace();
		}
    }
	
	private String getField(Element e, String key) {
		String field = null;
		Node n = null;
		
		if(e != null) {
			n = e.getElementsByTagName(key).item(0);
			
			if(n.getNodeType() == Node.ELEMENT_NODE) {
				e = (Element)n;
				
				field = getValue(e);
			}
		}
		
		return field;
	}
	
	private String getValue(Element e) {
		return e.getFirstChild().getNodeValue().trim();
	}
}
